﻿using System;
using Windows.Devices.Input;
using Windows.Foundation;
using Windows.UI;
using Windows.UI.Input;
using Windows.UI.Input.Inking;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Shapes;

namespace Chapter6
{
  public sealed partial class MainPage : Page
  {
    // Keeps track of the ink stroke data and can perform handwriting recognition
    InkManager inkManager = new InkManager();

    // Used for drawing strokes on the screen
    PathFigure currentFigure;

    public MainPage()
    {
      this.InitializeComponent();
      this.canvas.PointerPressed += Canvas_PointerPressed;
      this.canvas.PointerMoved += Canvas_PointerMoved;
      this.canvas.PointerReleased += Canvas_PointerReleased;
      this.canvas.PointerCanceled += Canvas_PointerReleased;
      this.canvas.PointerCaptureLost += Canvas_PointerReleased;
    }

    void Canvas_PointerPressed(object sender, PointerRoutedEventArgs e)
    {
      // Uncomment this if you want this sample to only work with pen input:
      // if (e.Pointer.PointerDeviceType == PointerDeviceType.Pen)
      {
        // Tell the InkManager about the PointerPressed event
        this.inkManager.ProcessPointerDown(e.GetCurrentPoint(this));

        // Set up a new shape to render the strokes between pointer down and up
        PathGeometry geometry = new PathGeometry();
        this.currentFigure = new PathFigure();
        geometry.Figures.Add(this.currentFigure);
        Path path = new Path { Data = geometry, StrokeThickness = 8, 
                               Stroke = new SolidColorBrush(Colors.Yellow) };

        // Add the (currently-empty) shape to the Canvas
        this.canvas.Children.Add(path);
        }
    }

    void Canvas_PointerMoved(object sender, PointerRoutedEventArgs e)
    {
      // Uncomment this if you want this sample to only work with pen input:
      // if (e.Pointer.PointerDeviceType == PointerDeviceType.Pen)
      {
        if (e.Pointer.IsInContact)
        {
          // Only do this when a pen is in use, and when it is making contact
          PointerPoint pointerPoint = e.GetCurrentPoint(this);

          // Tell the InkManager about the PointerMoved event.
          // Returns a Point when inking or selecting, or a Rect when erasing.
          object update = this.inkManager.ProcessPointerUpdate(pointerPoint);

          if (this.inkManager.Mode == InkManipulationMode.Inking)
          {
            DrawStroke((Point)update);
          }
          else if (this.inkManager.Mode == InkManipulationMode.Erasing)
          {
            Rect rect = (Rect)update;
            // The erasure is only successful if the invalidation Rect is nonzero
            if (rect.Height > 0 || rect.Width > 0)
            {
              RemoveErasedStrokes(rect);
            }
          }
        }
      }
    }

    void Canvas_PointerReleased(object sender, PointerRoutedEventArgs e)
    {
      // Uncomment this if you want this sample to only work with pen input:
      // if (e.Pointer.PointerDeviceType == PointerDeviceType.Pen)
      {
        // Tell the InkManager about the PointerPressed event
        this.inkManager.ProcessPointerUp(e.GetCurrentPoint(this));

        RecognizeStrokes();
      }
    }

    // Perform handwriting recognition and display all recognition guesses
    async void RecognizeStrokes()
    {
      listView.Items.Clear();

      foreach (InkRecognitionResult result in 
        await this.inkManager.RecognizeAsync(InkRecognitionTarget.All))
      {
        foreach (string candidate in result.GetTextCandidates())
        {
          // Display the candidate string
          listView.Items.Add(
            new ListViewItem { Content = candidate, FontSize = 50 });
        }
      }
    }

    void DrawStroke(Point point)
    {
      if (this.currentFigure.Segments.Count == 0)
      {
        // Set the starting point
        this.currentFigure.StartPoint = point;
      }

      // Add a line segment from the previous point to the current point
      this.currentFigure.Segments.Add(new LineSegment { Point = point });
    }

    void RemoveErasedStrokes(Rect invalidationRect)
    {
      // Just compare the segments from the current InkStrokes tracked by the
      // InkManager (which has at least one newly-erased stroke) to the segments
      // drawn on the screen, and remove any drawn segments that no longer
      // correspond to strokes managed by the InkManager.
      // Note that this ignores the passed-in invalidation rectangle.
      int pathIndex = 0, pathSegmentIndex = 0;
      Path currentPath = (Path)this.canvas.Children[pathIndex];
      PathSegmentCollection currentPathSegments = 
        ((PathGeometry)currentPath.Data).Figures[0].Segments;
      LineSegment currentPathSegment = 
        (LineSegment)currentPathSegments[pathSegmentIndex];

      foreach (InkStroke stroke in this.inkManager.GetStrokes())
      {
        foreach (InkStrokeRenderingSegment inkSegment in 
          stroke.GetRenderingSegments())
        {
          if (currentPathSegment.Point != inkSegment.Position)
          {
            // This has been erased, so remove it from the path
            currentPathSegments.Remove(currentPathSegment);

            // If this was the first segment in the path, 
            // update the StartPoint to match the new first segment
            if (pathSegmentIndex == 0 && currentPathSegments.Count > 0)
            {
              ((PathGeometry)currentPath.Data).Figures[0].StartPoint = 
                ((LineSegment)currentPathSegments[0]).Point;
            }

            // Account for the removed path segment
            pathSegmentIndex--;
          }

          // Update all the "current" references for the next iteration
          pathSegmentIndex++;
          if (pathSegmentIndex == currentPathSegments.Count)
          {
            // We've reached the end of this path
            pathIndex++;
            pathSegmentIndex = 0;
            if (pathIndex < this.canvas.Children.Count)
            {
              currentPath = (Path)this.canvas.Children[pathIndex];
              currentPathSegments = 
                ((PathGeometry)currentPath.Data).Figures[0].Segments;
              currentPathSegment = 
                (LineSegment)currentPathSegments[pathSegmentIndex];
            }
          }
        }
      }
    }
  }
}